@@ -8,12 +8,8 @@
 # or in the "LICENSE.txt" file accompanying this file. This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES
 # OR CONDITIONS OF ANY KIND, express or implied. See the License for the specific language governing permissions and
 # limitations under the License.
-# fmt: off
-from __future__ import absolute_import, print_function  # isort:skip
-from future import standard_library  # isort:skip
-standard_library.install_aliases()
-# fmt: on
 
+import functools
 import hashlib
 import json
 import logging
@@ -46,7 +42,7 @@ STACK_TYPE = "AWS::CloudFormation::Stack"
 class NodeType(Enum):
     """Enum that identifies the cluster node type."""
 
-    master = "Master"
+    head_node = "Master"
     compute = "Compute"
 
     def __str__(self):
@@ -138,8 +134,8 @@ def generate_random_name_with_prefix(name_prefix):
 
     Example: <name_prefix>-4htvo26lchkqeho1
     """
-    random_string = "".join(random.choice(string.ascii_lowercase + string.digits) for _ in range(16))
-    output_name = "-".join([name_prefix.lower()[: 63 - len(random_string) - 1], random_string])
+    random_string = "".join(random.choice(string.ascii_lowercase + string.digits) for _ in range(16))  # nosec
+    output_name = "-".join([name_prefix.lower()[: 63 - len(random_string) - 1], random_string])  # nosec
     return output_name
 
 
@@ -311,32 +307,15 @@ def upload_resources_artifacts(bucket_name, artifact_directory, root):
             bucket.upload_file(os.path.join(root, res), "%s/%s" % (artifact_directory, res))
 
 
-def get_instance_vcpus(instance_type, instance_info=None):
-    """
-    Get number of vcpus for the given instance type.
-
-    :param instance_type: the instance type to search for.
-    :return: the number of vcpus or -1 if the instance type cannot be found
-    """
-    try:
-        if not instance_info:
-            instance_info = get_instance_type(instance_type)
-
-        vcpus_info = instance_info.get("VCpuInfo")
-        vcpus = vcpus_info.get("DefaultVCpus")
-    except (ClientError):
-        vcpus = -1
-
-    return vcpus
-
-
 def get_supported_instance_types():
     """Return the list of instance types available in the given region."""
     ec2_client = boto3.client("ec2")
     try:
-        return [
+        supported_instance_types = set(
             offering.get("InstanceType") for offering in paginate_boto3(ec2_client.describe_instance_type_offerings)
-        ]
+        )
+        supported_instance_types.update(InstanceTypeInfo.additional_instance_types_data().keys())
+        return list(supported_instance_types)
     except ClientError as client_err:
         error(
             "Error when getting supported instance types via DescribeInstanceTypeOfferings: {0}".format(
@@ -576,20 +555,12 @@ def get_supported_os_for_scheduler(scheduler):
     :param scheduler: the scheduler for which we want to know the supported os
     :return: an array of strings of the supported OSes
     """
-    oses = ["alinux", "alinux2"]
+    oses = ["alinux2"]
     if scheduler != "awsbatch":
-        oses.extend(["centos7", "centos8", "ubuntu1604", "ubuntu1804"])
+        oses.extend(["centos7", "centos8", "ubuntu1804", "ubuntu2004"])
     return list(oses)
 
 
-def get_supported_os_for_architecture(architecture):
-    """Return list of supported OSes for the specified architecture."""
-    oses = ["alinux2", "ubuntu1804", "centos8"]
-    if architecture == "x86_64":
-        oses.extend(["centos7", "alinux", "ubuntu1604"])
-    return oses
-
-
 def get_supported_schedulers():
     """
     Return a tuple of the scheduler supported by parallelcluster.
@@ -759,12 +730,11 @@ def get_installed_version():
 def check_if_latest_version():
     """Check if the current package version is the latest one."""
     try:
-        latest = json.loads(urllib.request.urlopen("https://pypi.python.org/pypi/aws-parallelcluster/json").read())[
-            "info"
-        ]["version"]
+        pypi_url = "https://pypi.python.org/pypi/aws-parallelcluster/json"
+        latest = json.loads(urllib.request.urlopen(pypi_url).read())["info"]["version"]  # nosec nosemgrep
         if packaging.version.parse(get_installed_version()) < packaging.version.parse(latest):
             print("Info: There is a newer version %s of AWS ParallelCluster available." % latest)
-    except Exception:
+    except Exception:  # nosec
         pass
 
 
@@ -856,28 +826,28 @@ def describe_cluster_instances(stack_name, node_type):
     return instances
 
 
-def _get_master_server_ip(stack_name):
+def _get_head_node_ip(stack_name):
     """
-    Get the IP Address of the MasterServer.
+    Get the IP Address of the head node.
 
     :param stack_name: The name of the cloudformation stack
     :param config: Config object
     :return private/public ip address
     """
-    instances = describe_cluster_instances(stack_name, node_type=NodeType.master)
+    instances = describe_cluster_instances(stack_name, node_type=NodeType.head_node)
     if not instances:
-        error("MasterServer not running. Can't SSH")
-    master_instance = instances[0]
-    ip_address = master_instance.get("PublicIpAddress")
+        error("Head node not running. Can't SSH")
+    head_node = instances[0]
+    ip_address = head_node.get("PublicIpAddress")
     if ip_address is None:
-        ip_address = master_instance.get("PrivateIpAddress")
-    state = master_instance.get("State").get("Name")
+        ip_address = head_node.get("PrivateIpAddress")
+    state = head_node.get("State").get("Name")
     if state != "running" or ip_address is None:
         error("MasterServer: {0}\nCannot get ip address.".format(state.upper()))
     return ip_address
 
 
-def get_master_ip_and_username(cluster_name):
+def get_head_node_ip_and_username(cluster_name):
     cfn = boto3.client("cloudformation")
     try:
         stack_name = get_stack_name(cluster_name)
@@ -886,15 +856,15 @@ def get_master_ip_and_username(cluster_name):
         stack_status = stack_result.get("StackStatus")
 
         if stack_status in ["DELETE_COMPLETE", "DELETE_IN_PROGRESS"]:
-            error("Unable to retrieve master_ip and username for a stack in the status: {0}".format(stack_status))
+            error("Unable to retrieve head node ip and username for a stack in the status: {0}".format(stack_status))
         else:
-            master_ip = _get_master_server_ip(stack_name)
+            head_node_ip = _get_head_node_ip(stack_name)
             template = cfn.get_template(StackName=stack_name)
             mappings = template.get("TemplateBody").get("Mappings").get("OSFeatures")
             base_os = get_cfn_param(stack_result.get("Parameters"), "BaseOS")
             username = mappings.get(base_os).get("User")
 
-        if not master_ip:
+        if not head_node_ip:
             error("Failed to get cluster {0} ip.".format(cluster_name))
         if not username:
             error("Failed to get cluster {0} username.".format(cluster_name))
@@ -902,19 +872,19 @@ def get_master_ip_and_username(cluster_name):
     except ClientError as e:
         error(e.response.get("Error").get("Message"))
 
-    return master_ip, username
+    return head_node_ip, username
 
 
-def get_master_server_state(stack_name):
+def get_head_node_state(stack_name):
     """
-    Get the State of the MasterServer.
+    Get the State of the head node.
 
     :param stack_name: The name of the cloudformation stack
-    :return master server state name
+    :return head node state name
     """
     instances = describe_cluster_instances(stack_name, "Master")
     if not instances:
-        error("MasterServer not running.")
+        error("Head node not running.")
     return instances[0].get("State").get("Name")
 
 
@@ -939,20 +909,6 @@ def validate_pcluster_version_based_on_ami_name(ami_name):
             )
 
 
-def get_instance_types_info(instance_types, fail_on_error=True):
-    """Return InstanceTypes list returned by EC2's DescribeInstanceTypes API."""
-    try:
-        ec2_client = boto3.client("ec2")
-        return ec2_client.describe_instance_types(InstanceTypes=instance_types).get("InstanceTypes")
-    except ClientError as e:
-        error(
-            "Error when calling DescribeInstanceTypes for instances {0}: {1}".format(
-                ", ".join(instance_types), e.response.get("Error").get("Message")
-            ),
-            fail_on_error,
-        )
-
-
 def get_supported_architectures_for_instance_type(instance_type):
     """Get a list of architectures supported for the given instance type."""
     # "optimal" compute instance type (when using batch) implies the use of instances from the
@@ -961,8 +917,8 @@ def get_supported_architectures_for_instance_type(instance_type):
     if instance_type == "optimal":
         return ["x86_64"]
 
-    instance_info = get_instance_types_info([instance_type])[0]
-    supported_architectures = instance_info.get("ProcessorInfo").get("SupportedArchitectures")
+    instance_info = InstanceTypeInfo.init_from_instance_type(instance_type)
+    supported_architectures = instance_info.supported_architecture()
 
     # Some instance types support multiple architectures (x86_64 and i386). Filter unsupported ones.
     supported_architectures = list(set(supported_architectures) & set(SUPPORTED_ARCHITECTURES))
@@ -1021,7 +977,7 @@ def get_batch_ce(stack_name):
     """
     Get name of the AWS Batch Compute Environment.
 
-    :param stack_name: name of the master stack
+    :param stack_name: name of the head node stack
     :param config: config
     :return: ce_name or exit if not found
     """
@@ -1094,35 +1050,10 @@ def cluster_has_running_capacity(stack_name):
     return cluster_has_running_capacity.cached_result
 
 
-def get_instance_type(instance_type):
-    ec2_client = boto3.client("ec2")
-    try:
-        return ec2_client.describe_instance_types(InstanceTypes=[instance_type]).get("InstanceTypes")[0]
-    except Exception as e:
-        LOGGER.error("Failed when retrieving instance type data for instance type %s: %s", instance_type, e)
-        raise e
-
-
-def get_default_threads_per_core(instance_type, instance_info=None):
-    """Return the default threads per core for the given instance type."""
-    # NOTE: currently, .metal instances do not contain the DefaultThreadsPerCore
-    #       attribute in their VCpuInfo section. This is a known issue with the
-    #       ec2 DescribeInstanceTypes API. For these instance types an assumption
-    #       is made that if the instance's supported architectures list includes
-    #       x86_64 then the default is 2, otherwise it's 1.
-    if instance_info is None:
-        instance_info = get_instance_type(instance_type)
-    threads_per_core = instance_info.get("VCpuInfo", {}).get("DefaultThreadsPerCore")
-    if threads_per_core is None:
-        supported_architectures = instance_info.get("ProcessorInfo", {}).get("SupportedArchitectures", [])
-        threads_per_core = 2 if "x86_64" in supported_architectures else 1
-    return threads_per_core
-
-
 def disable_ht_via_cpu_options(instance_type, default_threads_per_core=None):
     """Return a boolean describing whether hyperthreading should be disabled via CPU options for instance_type."""
     if default_threads_per_core is None:
-        default_threads_per_core = get_default_threads_per_core(instance_type)
+        default_threads_per_core = InstanceTypeInfo.init_from_instance_type(instance_type).default_threads_per_core()
     res = all(
         [
             # If default threads per core is 1, HT doesn't need to be disabled
@@ -1156,7 +1087,7 @@ def read_remote_file(url):
             bucket, key = match.group(1), match.group(2)
             file_contents = boto3.resource("s3").Object(bucket, key).get()["Body"].read().decode("utf-8")
         else:
-            with urllib.request.urlopen(url) as f:
+            with urllib.request.urlopen(url) as f:  # nosec nosemgrep
                 file_contents = f.read().decode("utf-8")
         return file_contents
     except Exception as e:
@@ -1172,8 +1103,12 @@ def render_template(template_str, params_dict, tags, config_version=None):
     :param params_dict: Template parameters dict
     """
     try:
-        environment = Environment(loader=BaseLoader)
-        environment.filters["sha1"] = lambda value: hashlib.sha1(value.strip().encode()).hexdigest()
+        # A nosec comment is appended to the following line in order to disable the B701 check.
+        # This is done because it's needed to enable the desired functionality. The current callers
+        # of this function pass a template_str representing either a custom template specified by
+        # the user or the default template.
+        environment = Environment(loader=BaseLoader)  # nosec nosemgrep
+        environment.filters["sha1"] = lambda value: hashlib.sha1(value.strip().encode()).hexdigest()  # nosec nosemgrep
         environment.filters["bool"] = lambda value: value.lower() == "true"
         template = environment.from_string(template_str)
         output_from_parsed_template = template.render(config=params_dict, config_version=config_version, tags=tags)
@@ -1225,25 +1160,208 @@ def get_ebs_snapshot_info(ebs_snapshot_id, raise_exceptions=False):
         )
 
 
-def get_instance_network_interfaces(instance_type, instance_info=None):
-    """Return the number of network interfaces to configure for the instance type."""
-    if not instance_info:
-        instance_info = get_instance_type(instance_type)
-
-    # Until maximumNetworkCards is not available, 1 is a safe value for all instance types
-    needed_interfaces = int(instance_info.get("NetworkInfo").get("MaximumNetworkCards", 1))
-
-    return needed_interfaces
+def get_default_instance_type():
+    """If current region support free tier, return the free tier instance type. Otherwise, return t3.micro ."""
+    if not hasattr(get_default_instance_type, "cache"):
+        get_default_instance_type.cache = {}
+    cache = get_default_instance_type.cache
+    region = os.environ.get("AWS_DEFAULT_REGION")
+    if region not in cache:
+        free_tier_instance_type = []
+        for page in paginate_boto3(
+            boto3.client("ec2").describe_instance_types,
+            Filters=[
+                {"Name": "free-tier-eligible", "Values": ["true"]},
+                {"Name": "current-generation", "Values": ["true"]},
+            ],
+        ):
+            free_tier_instance_type.append(page)
+        cache[region] = free_tier_instance_type[0]["InstanceType"] if free_tier_instance_type else "t3.micro"
+    return cache[region]
 
 
-def get_instance_gpus(instance_type, instance_info=None):
-    """Return the number of GPUs provided by the instance type."""
-    if not instance_info:
-        instance_info = get_instance_type(instance_type)
+class Cache:
+    """Simple utility class providing a cache mechanism for expensive functions."""
 
-    gpu_info = instance_info.get("GpuInfo", None)
+    _caches = []
 
-    # Currently adding up all gpus. To be reviewed if the case of heterogeneous GPUs arises.
-    gpus = sum([gpus.get("Count") for gpus in gpu_info.get("Gpus")]) if gpu_info else 0
+    @staticmethod
+    def is_enabled():
+        """Tell if the cache is enabled."""
+        return not os.environ.get("PCLUSTER_CACHE_DISABLED")
 
-    return gpus
+    @staticmethod
+    def clear_all():
+        """Clear the content of all caches."""
+        for cache in Cache._caches:
+            cache.clear()
+
+    @staticmethod
+    def _make_key(args, kwargs):
+        key = args
+        if kwargs:
+            for item in kwargs.items():
+                key += item
+        return hash(key)
+
+    @staticmethod
+    def cached(function):
+        """
+        Decorate a function to make it use a results cache based on passed arguments.
+
+        Note: all arguments must be hashable for this function to work properly.
+        """
+        cache = {}
+        Cache._caches.append(cache)
+
+        @functools.wraps(function)
+        def wrapper(*args, **kwargs):
+            cache_key = Cache._make_key(args, kwargs)
+
+            if Cache.is_enabled() and cache_key in cache:
+                return cache[cache_key]
+            else:
+                return_value = function(*args, **kwargs)
+                if Cache.is_enabled():
+                    cache[cache_key] = return_value
+                return return_value
+
+        return wrapper
+
+
+class InstanceTypeInfo:
+    """
+    Data object wrapping the result of a describe_instance_types call.
+
+    TODO: This class stores additional instance types data globally. In the future we should have a separate instance
+    per cluster configuration.
+    """
+
+    # Additional data used to describe instance types
+    __additional_instance_types_data = {}
+
+    def __init__(self, instance_type_data):
+        self.instance_type_data = instance_type_data
+
+    @staticmethod
+    def load_additional_instance_types_data(instance_types_data):
+        """Load additional data to describe instance types."""
+        InstanceTypeInfo.__additional_instance_types_data = instance_types_data if instance_types_data else {}
+
+    @staticmethod
+    def clear_additional_instance_types_data():
+        """Clear the additional data used to describe instance types."""
+        InstanceTypeInfo.__additional_instance_types_data = {}
+
+    @staticmethod
+    def additional_instance_types_data():
+        """Get the additional data used to describe instance types."""
+        return InstanceTypeInfo.__additional_instance_types_data
+
+    @staticmethod
+    @Cache.cached
+    def init_from_instance_type(instance_type, exit_on_error=True):
+        """
+        Init InstanceTypeInfo by performing a describe_instance_types call.
+
+        Multiple calls for the same instance_type are cached.
+        The function exits with error if exit_on_error is set to True.
+        """
+        try:
+            if instance_type in InstanceTypeInfo.additional_instance_types_data():
+                instance_type_data = InstanceTypeInfo.additional_instance_types_data()[instance_type]
+            else:
+                ec2_client = boto3.client("ec2")
+                instance_type_data = ec2_client.describe_instance_types(InstanceTypes=[instance_type]).get(
+                    "InstanceTypes"
+                )[0]
+            return InstanceTypeInfo(instance_type_data)
+        except ClientError as e:
+            error(
+                "Failed when retrieving instance type data for instance {0}: {1}".format(
+                    instance_type, e.response.get("Error").get("Message")
+                ),
+                exit_on_error,
+            )
+
+    def gpu_count(self):
+        """Return the number of GPUs for the instance."""
+        gpu_info = self.instance_type_data.get("GpuInfo", None)
+
+        gpu_count = 0
+        if gpu_info:
+            for gpus in gpu_info.get("Gpus", []):
+                gpu_manufacturer = gpus.get("Manufacturer", "")
+                if gpu_manufacturer.upper() == "NVIDIA":
+                    gpu_count += gpus.get("Count", 0)
+                else:
+                    warn(
+                        "ParallelCluster currently does not offer native support for '{0}' GPUs. "
+                        "Please make sure to use a custom AMI with the appropriate drivers in order to leverage "
+                        "GPUs functionalities".format(gpu_manufacturer)
+                    )
+
+        return gpu_count
+
+    def gpu_type(self):
+        """Return name or type of the GPU for the instance."""
+        gpu_info = self.instance_type_data.get("GpuInfo", None)
+        # Remove space and change to all lowercase for name
+        return "no_gpu_type" if not gpu_info else gpu_info.get("Gpus")[0].get("Name").replace(" ", "").lower()
+
+    def max_network_interface_count(self):
+        """Max number of NICs for the instance."""
+        needed_interfaces = int(self.instance_type_data.get("NetworkInfo").get("MaximumNetworkCards", 1))
+        return needed_interfaces
+
+    def default_threads_per_core(self):
+        """Return the default threads per core for the given instance type."""
+        # NOTE: currently, .metal instances do not contain the DefaultThreadsPerCore
+        #       attribute in their VCpuInfo section. This is a known issue with the
+        #       ec2 DescribeInstanceTypes API. For these instance types an assumption
+        #       is made that if the instance's supported architectures list includes
+        #       x86_64 then the default is 2, otherwise it's 1.
+        threads_per_core = self.instance_type_data.get("VCpuInfo", {}).get("DefaultThreadsPerCore")
+        if threads_per_core is None:
+            supported_architectures = self.instance_type_data.get("ProcessorInfo", {}).get("SupportedArchitectures", [])
+            threads_per_core = 2 if "x86_64" in supported_architectures else 1
+        return threads_per_core
+
+    def vcpus_count(self):
+        """Get number of vcpus for the given instance type."""
+        try:
+            vcpus_info = self.instance_type_data.get("VCpuInfo")
+            vcpus = vcpus_info.get("DefaultVCpus")
+        except ClientError:
+            vcpus = -1
+
+        return vcpus
+
+    def supported_architecture(self):
+        """Return the list of supported architectures."""
+        return self.instance_type_data.get("ProcessorInfo").get("SupportedArchitectures")
+
+    def is_efa_supported(self):
+        """Check whether EFA is supported."""
+        return self.instance_type_data.get("NetworkInfo").get("EfaSupported")
+
+    def supported_usage_classes(self):
+        """Return the list supported usage classes."""
+        supported_classes = list(self.instance_type_data.get("SupportedUsageClasses", []))
+        if "on-demand" in supported_classes:
+            # Replace official AWS with internal naming convention
+            supported_classes.remove("on-demand")
+            supported_classes.append("ondemand")
+        return supported_classes
+
+
+@Cache.cached
+def get_fsx_info(fs_id):
+    try:
+        return boto3.client("fsx").describe_file_systems(FileSystemIds=[fs_id]).get("FileSystems")[0]
+    except ClientError as e:
+        error(
+            "Failed when retrieving FSx filesystem data for filesystem {0}: {1}".format(
+                fs_id, e.response.get("Error").get("Message")
+            )
+        )
